#! /usr/bin/perl 
#This program is used to clustering 16S rRNA sequences at species level.
#
# Last Modified: NOV 26, 2012
###########################################################################
# usage: perl Mpick.pl -s sequencefilename(with path) [-e epsilon] [-d delta] [-f (0 or 1)]
#	Parameters:
#		-s sequence input file
#       -e \epsilon, parameter to form a graph
#       -d \delta, parameter used for stopping criterion
#       -f 1:fast implement with uclust preprocessing; 0:regular implement.
#   Output:
#       clustering result in text format
#   Example:
#       perl Mpick.pl -s ../test/test.fas -f 1
###########################################################################



use strict;
use warnings;
use diagnostics;


my $size = scalar @ARGV;
my $allRequired = 1;

my ($inputSeqName,$epsilon,$delta,$fast);
$fast=0;
for(my $a = 0; $a < $size; $a++)
{
	if($ARGV[$a] eq "-s")
	{
		$a++;
		$allRequired--;
		$inputSeqName = $ARGV[$a];
	}
	elsif($ARGV[$a] eq "-d")
	{
		$a++;
		$delta = $ARGV[$a];
	}
    elsif($ARGV[$a] eq "-e")
	{
		$a++;
		$epsilon = $ARGV[$a];
	}
    elsif($ARGV[$a] eq "-f")
	{
		$a++;
		$fast = $ARGV[$a];
	}
}

if($allRequired != 0)
{
	print "Please make sure you entered the sequence file.\n";
	exit;
}

my $ind=rindex($inputSeqName,"\.");
my $pre=substr($inputSeqName,0,$ind); #the file name without extention.

my $numnode=0; #number of sequences in a sequence file
open(FILEseq,$inputSeqName) or die "can't open $inputSeqName";
my @fas=<FILEseq>;
close FILEseq;
for my $i (0 .. $#fas)
{
    my $line=$fas[$i];
    if ($line =~ />/)
    {
        $numnode++;
    }
}

my @otu_seq=(); #otu-sequence map for fast implement
my ($pre1,$inputSeqName1); #preface of fast
if ($fast ==1) #fast implement?
{
    print "Preprocessing with UCLUST...\n";
    my $cmd="./uclust.sh $inputSeqName 1";
    system($cmd);
    
    my $uclustfile=$pre.".ucluster"; #the clusters generated from uclust
    open(FILEfas,$inputSeqName) or die "can't open $inputSeqName";
    my @fas=<FILEfas>;
    close FILEfas;
    my @ uclustinclude=();
    for my $i (0 ..$numnode-1)
    {
        $ uclustinclude[$i]=0; #not included in the uclust result, initialization.
    }
    
    my @seqfas=();
    my @seqlab=();
    my $seqfas="";
    my $countfas=0;
    for (my $i =0; $i<=$#fas; $i++)
    {
        my $line=$fas[$i];
        $line=~s/\s+$//;
        if ($line =~ />/)
        {
            $seqlab[$countfas]=$line;
            if (length($seqfas)>0)
            {
                $seqfas[$countfas-1]=$seqfas;
                $seqfas="";
            }
            $countfas++;
        }else
        {
            $seqfas.=$line;
        }
    }
    $seqfas[$countfas-1]=$seqfas;
    
    open(FILEclust,$uclustfile) or die "can't open $uclustfile";
    my @clust=<FILEclust>;
    close FILEclust;
    my $line=$clust[0];
    $line=~s/\s+$//;
    @otu_seq=split(/\|/,$line); #otus from uclust at 99%
    for my $i (0 ..$#otu_seq)
    {
        my $line1=$otu_seq[$i];
        $line1=~s/^\s+//;
        $line1=~s/\s+$//;
        my @line1=split(/\s+/,$line1);
        for my $j (0..$#line1)
        {
            my $line2=$line1[$j];
            $ uclustinclude[$line2]=1;
        }
    }
    
    for my $i (0 .. $numnode-1)
    {
        if ($uclustinclude[$i]==0) # the seq is not included in the uclust result, add it.
        {
            $otu_seq[$#otu_seq+1]=$i;
        }
    }
    my @rep_otu=();
    for (my $i=0; $i<=$#otu_seq; $i++)
    {
        my $line=$otu_seq[$i];
        $line=~s/^\s+//;
        $line=~s/\s+$//;
        my @line=split(/\s+/,$line);
        my $maxfaslen=0;
        for (my $j=0; $j<=$#line; $j++)
        {
            my $len=length($seqfas[$line[$j]]);
            unless (defined($len))
            {
                my $nop=1;
            }
            if ($len ==0)
            {
                my $nop=1;
            }
            if ($len>$maxfaslen)
            {
                $maxfaslen=$len;
                $rep_otu[$i]=$line[$j];
            }
        }
    }
    
    my $uclustfas=$pre."_uclust.fas";
    open(FILEout,">$uclustfas") or die "can't open $uclustfas";
    for (my $i=0; $i<=$#otu_seq; $i++)
    {
        print FILEout "$seqlab[$rep_otu[$i]]\n";
        print FILEout "$seqfas[$rep_otu[$i]]\n";
    }
    close(FILEout);
    $inputSeqName1=$pre."_uclust.fas";
    $pre1=$pre."_uclust";
}



my $smallestcluster=5;
unless(defined($delta))
{
    $delta=0.1; #default
}
unless(defined($epsilon))
{
    $epsilon=0.04; #default
}

my $kmer=0; #not use kmer

my $ndistfilesize=0;
if (-e $pre.".ndist")
{
    $ndistfilesize= -s $pre.".ndist";
}
if ($ndistfilesize < 1000) #not to compute the distances if the distance file exits
{
	print "Computing pairwise distances...It may take some time...\n";
    unless ($fast ==1)
    {
        &create_dist($inputSeqName);
        system("rm $pre"."*.dist");
        system("rm $pre"."*.ndist_*");
    }else
    {
        &create_dist($inputSeqName1);
        system("rm $pre1"."*.dist");
        system("rm $pre1"."*.ndist_*");
        system("rm $pre1".".fas");
        system("rm $pre".".ucluster");
        my $ndistfile=$pre.".ndist";
        open(FILEout,">$ndistfile") or die "can't open $ndistfile";
        my $fastndistfile=$pre1.".ndist";
        open(FILEin,$fastndistfile) or die "can't open $fastndistfile";
        my @fastdist=<FILEin>;
        close(FILEin);
        for my $i (0..$#fastdist) #seqs in different otus
        {
            my $line=$fastdist[$i];
            $line=~s/^\s+//;
            $line=~s/\s+$//;
            my @line=split(/\s+/,$line);
            my $seq1=$line[0];
            my $otuseq1=$otu_seq[$seq1];
            $otuseq1=~s/^\s+//;
            $otuseq1=~s/\s+$//;
            my $seq2=$line[1];
            my $otuseq2=$otu_seq[$seq2];
            $otuseq2=~s/^\s+//;
            $otuseq2=~s/\s+$//;
            my $dist=$line[2];
            my @otuseq1=split(/\s+/,$otuseq1);
            my @otuseq2=split(/\s+/,$otuseq2);
            for my $j (0..$#otuseq1)
            {
                for my $k (0..$#otuseq2)
                {
                    if ($otuseq1[$j]<$otuseq2[$k])
                    {
                        print FILEout $otuseq1[$j]."\t".$otuseq2[$k]."\t".$dist."\n";
                    }else
                    {
                        print FILEout $otuseq2[$k]."\t".$otuseq1[$j]."\t".$dist."\n";
                    }
                }
            }
        }
        for my $i (0 .. $#otu_seq) #seqs in the same otu
        {
            my $line=$otu_seq[$i];
            $line=~s/^\s+//;
            $line=~s/\s+$//;
            my @line=split(/\s+/,$line);
            for my $j (0..$#line)
            {
                for my $k ($j+1 .. $#line)
                {
                    if ($line[$j]<$line[$k])
                    {
                       print FILEout $line[$j]."\t".$line[$k]."\t"."0.005"."\n";
                    }else
                    {
                        print FILEout $line[$k]."\t".$line[$j]."\t"."0.005"."\n";
                    }
                }
            }
        }
        close(FILEout);
        system("rm $pre1".".ndist");
        system("rm $pre1".".fas");
    }
}


my $goon=1; #recursively run the clustering if goon=1
my @checkres=();
$checkres[0]=0;
&cluster_mpick($pre);


sub cluster_mpick { #input $pre. M-pick main function
	
	my ($pre)=@_;

	print "Clustering begins...\n";
    my $cmd="";
    if (-e $pre.".mpick")
    {
        $cmd="rm $pre".".mpick";
        system($cmd);
    }
    print "Generating graph. epsilon=$epsilon\n";
	$cmd="perl gentree.pl -t $pre".".ndist -d $epsilon -n $numnode"; #generate tree file for clustering
    system($cmd);
	$cmd="./convert -i $pre".".tree -o $pre".".bin -w";
	system($cmd);

	my $t=1; #just use t=1;
	$cmd="./community $pre".".bin -l -1 -w -t $t > $pre".".dat "; #do modularity clustering
	system($cmd);
	$cmd="perl genclusters.pl -d $pre".".dat -i 1 -n $numnode";  #get the first result
	system($cmd);

	my $clusterfile="$pre".".mpick";
	my $distfile="$pre".".ndist";
	open (Distfile, $distfile) or die "can't open $distfile";
	my @distfile=<Distfile>;
	close Distfile;
	my @distarr=();
	for (my $i=0; $i<$numnode; $i++)
	{
		for (my $j=$i+1; $j<$numnode; $j++)
		{
			$distarr[$i][$j]=$epsilon;
			
		}
	}
	for (my $i=0; $i<=$#distfile; $i++)
	{
		my $distline=$distfile[$i];
		if (length($distline)>0)
		{
			my @distline=split(/\s+/,$distline);
			my $seq1=$distline[0];
			my $seq2=$distline[1];
			if ($distline[2]<=$epsilon && $distline[2]>=0)
			{
				$distarr[$seq1][$seq2]=$distline[2];
			}
		}
	}
	my $refdistarr=\@distarr;
	
	my $level=1;
	print "Starting recursive clustering...\n";
	while ($goon == 1)
	{
		print "Level: $level \n";
		&nextlevel($clusterfile,$refdistarr,$level,$numnode,$epsilon);
        $level++;
	}
    &clean_files($pre);
}

sub nextlevel {#input: clusterfile, distfile, level. output: clusterfile. to carry out M-pick on the next level.
	my ($clusterfile,$refdistarr,$level,$numnode,$epsilon)=@_;
    open (Clustfile, $clusterfile) or die "can't open $clusterfile";
	my @clusterfile=<Clustfile>;
	close Clustfile;
	
	my @nextlevel=();
	my $clusters=$clusterfile[$level-1];
	my @clusters=split(/\|/,$clusters);
	
	my $newlevel=$level+1;
	$nextlevel[0]="$newlevel |";
	$goon=0;
    my @precheckres=();
	my @nextcheckres=();
		
	for (my $i=1; $i<$#clusters; $i++)
	{
		if ($level==1)
		{
			$precheckres[$i]=0; #initialize
		}else
		{
            $precheckres[$i]=$checkres[$i];
		}
	}
	
	
	for (my $i=1; $i<$#clusters; $i++) #look into each cluster to see if it should be further partitioned. 
	{
		my $cluster=$clusters[$i];
		my @cluster=split(/\s+/,$cluster);
		my @sort_cluster=sort { $a <=> $b } @cluster;
		my $sort_cluster=join(" ",@sort_cluster); #node indices in a cluster need to be sorted to suite the structure of distancematrix.
		
		if ($#cluster <=$smallestcluster)
		{
			$precheckres[$i]=1;
		}
		
		
		if ($precheckres[$i]==0) #do further partitions.
		{
			my $tmptree=$pre."_$level"."_$i.tmptree";
			my $mindist=1000; #minimum distance
			for (my $j=0; $j<=$#sort_cluster; $j++)
			{
				for (my $k=$j+1; $k<=$#sort_cluster; $k++)
				{
					my $seq1=$sort_cluster[$j];
					my $seq2=$sort_cluster[$k];
					my $dist=&roundint(($epsilon-${$refdistarr}[$seq1][$seq2])*100);
					if ($dist<$mindist && $dist>=0)
					{
						$mindist=$dist;
					}
				}
			}
			if ($mindist <=0)			{
				$mindist=1;
			}
			
			my $tmptreec="";
			for (my $j=0; $j<=$#sort_cluster; $j++)
			{
				for (my $k=$j+1; $k<=$#sort_cluster; $k++)
				{
					my $seq1=$sort_cluster[$j];
					my $seq2=$sort_cluster[$k];
					my $dist=&roundint(($epsilon-${$refdistarr}[$seq1][$seq2])*100)-$mindist+1; #form the modified similarity
					if ($dist>0)
					{
						$tmptreec.="$seq1\t$seq2\t$dist\n";
					}
				}
			}
			
			if (length($tmptreec)>0)
			{
				open (Tmptreefile,">$tmptree") or die "can't open $tmptree";
				print Tmptreefile $tmptreec;
				close Tmptreefile;
			}

            my $cmd="./convert -i $tmptree -o $pre"."_$level"."_$i".".tmpbin -w";
			system($cmd);
			my $t=1; #just use t=1;
			$cmd="./community $pre"."_$level"."_$i".".tmpbin -l -1 -w -t $t > $pre"."_$level"."_$i".".tmpdat ";
			system($cmd);
			my $qfile="$pre"."_$level"."_$i".".tmpbin.txt";
			open(FILEq,$qfile) or die "can't open $qfile";
			my @q=<FILEq>;
			if ($q[0]>$delta) #modularity is greater than delta, need to be further partitionned.
			{
				$goon=1;
				my $num_tmpnode=$#cluster+1;
				if (-e "$pre"."_$level"."_$i".".tmpmpick")
				{
					system("rm $pre"."_$level"."_$i".".tmpmpick");
				}
				$cmd="perl genclusters.pl -d $pre"."_$level"."_$i".".tmpdat -i 1 -n $num_tmpnode -o $pre"."_$level"."_$i".".tmpmpick";
				system($cmd);
			
				$nextlevel[$i]=&transformlevel("$pre"."_$level"."_$i".".tmpmpick",$sort_cluster);
				my $testlevel=$nextlevel[$i];
				$testlevel=~s/\s+$//;
				$testlevel=~s/\|$//;
				my @testlevel=split(/\|/,$testlevel);
				for (my $j=0; $j<=$#testlevel; $j++)
				{
					push(@nextcheckres,0); #need to check in the next run
				}
				unless ($testlevel =~/\|/) #no further split appears in the result
				{
					$goon=0;
				}
			
				system("rm $pre"."_$level"."_$i".".tmpmpick");
			}else
			{
				$nextlevel[$i]="$sort_cluster|";
				push(@nextcheckres,1); #doesn't need to be split any more
			}
        
		}else
		{
			$nextlevel[$i]="$sort_cluster|";
			push(@nextcheckres,1);
		}
		
	}
	open (Clustfile, ">>$clusterfile") or die "can't open $clusterfile";
	for (my $i=0; $i<$#clusters; $i++)
	{
		print Clustfile $nextlevel[$i];
	}
	print Clustfile "\n";
	close Clustfile;
	
	for (my $i=0; $i<=$#nextcheckres; $i++)
	{
        $checkres[$i+1]=$nextcheckres[$i]; #update the checkres
	}
	
}

sub transformlevel { #input: datfile, origianl seq numbers in a cluster. output: transform back to the original sequence index numbers in the new level
	my ($datfile,$cluster)=@_;
	my $rescluster="";
	my @rescluster=();
	my @cluster=split(/\s+/,$cluster);
	open (FILEdat, $datfile) or die "can't open $datfile";
	my @dat=<FILEdat>;
	close FILEdat;
	my $dat=$dat[0];
	@dat=split(/\|/,$dat);
	for (my $i=1; $i<$#dat; $i++)
	{
		my $dat_cluster=$dat[$i];
		my @dat_cluster=split(/\s+/,$dat_cluster);
		$rescluster[$i]="";
		my @tmpcluster=();
		for (my $j=0; $j<=$#dat_cluster; $j++)
		{
			push(@tmpcluster,$cluster[$dat_cluster[$j]]);
		}
		$rescluster[$i]=join(" ",@tmpcluster);
		$rescluster=$rescluster.$rescluster[$i]."|";

	}
	return $rescluster;
}

sub clean_files { #input: $pre
    my ($pre)=$pre;
	my $cmd="rm $pre".".tree";
	system($cmd);
	$cmd="rm $pre"."*.tmptree";
	system($cmd);
    $cmd="rm $pre"."*.bin";
	system($cmd);
	$cmd="rm $pre"."*.tmpbin";
	system($cmd);
    $cmd="rm $pre"."*.tmpbin.txt";
	system($cmd);
    $cmd="rm $pre"."*.bin.txt";
	system($cmd);
    $cmd="rm $pre"."*.dat";
	system($cmd);
	$cmd="rm $pre"."*.tmpdat";
    system($cmd);
}

sub create_dist { #input $pre. compute sequence distances
	my ($inputSeqName)=@_;
	my $cmd="perl MT_esprit_needle.pl -s $inputSeqName -n 5"; #-n: using n(n+1)/2 threads to compute the distances
	system($cmd);
}


sub roundint { #input: int1. output:roundint
	my ($num)=@_;
	my $roundint=0;
	if ($num-int($num)<=0.5)
	{
		$roundint=int($num);
	}else
	{
		$roundint=int($num)+1;
	}
	return $roundint;
}
