#! /usr/bin/perl 
#This program is used to generate a formatted cluster result from the modularity clustering result
#
# Last Modified: SEP 26, 2012
###########################################################################
# usage: perl genclusters.pl -d datafile -i ID -n numnode -o outputfile
#	Parameters:
#       -d datfile generated by modularity clustering
#		-i index of clustering result
#       -n numnode, number of nodes
#   Output:
#       -o final clustering result
#   Example:
#       perl genclusters.pl -d ../test/test.data -i 1 -n 110 -o ../test/test.mpick1
###########################################################################

use strict;
use warnings;
use diagnostics;
#use threads;

my ($datfile,$ID,$numnode,$output);
my $size = scalar @ARGV;
for(my $a = 0; $a < $size; $a++)
{
	if($ARGV[$a] eq "-d")
	{
		$a++;
		$datfile = $ARGV[$a];
	}
	elsif($ARGV[$a] eq "-i")
	{
		$a++;
		$ID = $ARGV[$a];
	}
    elsif($ARGV[$a] eq "-n")
	{
		$a++;
		$numnode = $ARGV[$a];
	}
    elsif($ARGV[$a] eq "-o")
	{
		$a++;
		$output = $ARGV[$a];
	}
}

$datfile=~/(\S+)\.(\w+)$/;
my $pre=$1;
unless(defined($output))
{
    $output=$pre.".mpick";
}

my $addlevel=-1;
open(FILEtree,$datfile) or die "can't open $datfile";
my @data1=<FILEtree>;
close FILEtree;
my @data=();
my $count=0;
for my $i (0..$#data1)
{
    my $line=$data1[$i];
    if ($line=~/^\d+/)
    {
        $data[$count]=$line;
        $count++;
    }
}
my @seclev=(); #keep the second level cluster info
my $seclev_start=-1;

if ($#data>$numnode)
{
	my $datalast=$data[$numnode];
	my @nodelast=split(/\s+/,$datalast);

	my $nodelast=$nodelast[0];

	if ($nodelast == $numnode)
	{
#the split distance is small, so unique links are added
#to make the program work, it couldn't use self link. I added a node 500, and link nodes to it.

		$addlevel=$nodelast[1];
	}
	if (($#data-$numnode)> 3)
	{
		if ($addlevel>0)
		{
			$seclev_start=$numnode+1;
		}else
		{
			$seclev_start=$numnode;
		}
	}
	
}

if ($seclev_start >0) #has second level structure
{
	my $oldcluster=$data[$seclev_start];
	my @oldcluster=split(/\s+/,$oldcluster);
	$oldcluster=$oldcluster[0];
	my $flag=0; #sign of stop counting
	my $k=0;
	for (my $i=$seclev_start; $i<=$#data; $i++)
	{
		my $newcluster=$data[$i];
		if (length($newcluster)>1) #not blank line
		{
			my @newcluster=split(/\s+/,$newcluster);
			$newcluster=$newcluster[0];
			my $newcluster1=$newcluster[1];
			if ($flag ==0)
			{
				if ($newcluster>=$oldcluster)
				{
					$seclev[$k]=$newcluster1;
					$k++;
					$oldcluster=$newcluster;
				}else
				{
					$flag=1; #stop
				}
			}
		}
	}
}

	
my @res=(); # $res[0]=22 23...
my $numlevel=0; #max num of levels
for (my $i=0; $i<$numnode; $i++)
{
	my @line=split(/\s+/,$data[$i]);
	my $line1=$line[1];
	if ($line1>$numlevel)
	{
		$numlevel=$line1;
	}
}

my $k=0; #level count
for (my $i=0; $i<=$numlevel; $i++)
{
	my $res="";
	if ($i != $addlevel)
	{
		for (my $j=0; $j<$numnode; $j++)
		{
			my $line=$data[$j];
			my @line=split(/\s+/,$line);
			my $line1=$line[1];
		
			if ($line1 == $i)
			{
				$res=$res.$line[0]." ";
			}
		}
		$res=~s/\s+$//;
		$res[$k]=$res;
		$k++;
	}
}

my @res1=();
if ($#seclev>0)
{
	for (my $i=0; $i<=$#seclev; $i++)
	{
		my $j=$seclev[$i];
		if (defined($res1[$j]))
		{
			$res1[$j]=$res1[$j]." ".$res[$i];
		}else
		{
			$res1[$j]=$res[$i];
		}
	}
}else
{
	@res1=@res;
}

if ($addlevel >0)
{
	$k=$#res1;
}else
{
	$k=$#res1+1;
}
if ($addlevel >0)
{
	for (my $j=0; $j<$numnode; $j++)
	{
		my $line=$data[$j];
		my @line=split(/\s+/,$line);
		my $line1=$line[1];
		if ($line1 == $addlevel)
		{
			$res1[$k]=$line[0];
			$k++;
		}
	}
}

open (FILEout,">>$output") or die "can't open $output";

print FILEout "$ID |";
for (my $i=0; $i<=$#res1; $i++)
{
	print FILEout "$res1[$i]|";
	
}
print FILEout "\n";
close FILEout;

